# -*- coding: utf-8 -*-
from luckydonaldUtils.logger import logging

try:
    from .pip_interface import pip_install
except ImportError as e:
    logger = logging.getLogger(__name__)
    logger.warn("pip_interface: Could not apply logger workaround. Falling back to using pip as import.")
    try:
        from .pip_interface_fallback import pip_install
    except ImportError as e:
        logger.warn("pip_interface: Could not import pip module. Falling back to using pip directly via cli call.")
        from .pip_interface_fallback_fallback import pip_install
    # end try
# end try

from setuptools import find_packages

try:
    import importlib
except ImportError:
    # pip install importlib
    pip_install(["importlib"])
    import importlib
# end try

__author__ = 'luckydonald'
logger = logging.getLogger(__name__)


def import_or_install(package_name, pip_name=None):
    """
    Tries to import a package.
    If that fails, it tries to install it via pip, using the given `pip_name` or if not given, the `package_name`.
    :param package_name:  Package name to import. (E.g. "PIL")
    :param pip_name:  The name to install it like `$pip install <pip_name>` would do. (E.g. "Pillow")
    :return:
    """
    # if pip name is given, just use that.
    if pip_name:
        return import_or_install_with_exact_pip_name(package_name, pip_name)
    # if pip name is not given, try possibilities, by splitting the dots.
    #
    # Example:
    #  >> import_or_install("imgurpython.client.ImgurClient")
    # will try to install:
    #   - "imgurpython"
    #   - "imgurpython.client"
    #   - "imgurpython.client.ImgurClient"
    #
    # This also allows package names containing dots like "ruamel.yaml".
    pip_name = ""
    err = None
    for part in package_name.split("."):
        pip_name = (pip_name + "." if pip_name else "") + part
        try:
            return import_or_install_with_exact_pip_name(package_name, pip_name)
        except ImportError as e:
            err = e
            logger.debug("Import failed.", exc_info=True)
            # end try
    # end for
    raise err  # should store the last occurred error.


# end def


def import_or_install_with_exact_pip_name(package_name, pip_name):
    """
    Just a helper for import_or_install()

    Also, Littlepip is best pony.
    """
    err = None
    for try_i in [1, 2, 3]:
        try:
            return import_only(package_name)
        except ImportError as e:
            err = e
            logger.debug("Import failed.", exc_info=True)
            upgrade = try_i >= 2  # import failed twice (we first try normal install and only if that doesn't work, an --upgrade)
            install_only(pip_name, upgrade)
    raise err  # should store the last occurred error.


# end def


def import_only(package_name, module_list=None):
    # "pytz.timzone" -> from pytz import timezone -> package_name = "pytz", from_package = ["timezone"]
    if not module_list:
        if "." in package_name:
            package_name, module_list = package_name.rsplit('.', 1)
        else:
            module_list = None
    if module_list:
        logger.debug("Trying import: form \"{module_name}\" import \"{module_list}\".".format(module_name=package_name,
                                                                                              module_list=module_list))
    else:
        logger.debug("Trying to import module \"{module_name}\".".format(module_name=package_name))

    try:
        imp = importlib.import_module(package_name)
        if module_list:
            if hasattr(imp, module_list):
                imp = getattr(imp, module_list)
                logger.debug("\"{module_list}\" is an attribute of \"{module_name}\".".format(module_name=package_name,
                                                                                              module_list=module_list))
            else:
                imp = importlib.import_module(package_name, package=module_list)
                logger.debug("\"{module_list}\" is an module in \"{module_name}\".".format(module_name=package_name,
                                                                                           module_list=module_list))
        else:
            imp = importlib.import_module(package_name, package=module_list)
            logger.debug("module \"{module_name}\".".format(module_name=package_name))

    except ImportError:
        try:
            imp = importlib.import_module(package_name, package=module_list)
        except (SystemError, ValueError) as e:
            # https://github.com/luckydonald/luckydonald-utils/issues/2
            # https://bugs.python.org/issue18018
            raise ImportError(str(e))
    return imp


def install_only(pip_name, upgrade=False):
    logger.warn("{install_or_upgrade} package '{pip_name}'.\n"
                "If that fails, install it manually:\n"
                "pip install {pip_name}\n"
                "".format(pip_name=pip_name, install_or_upgrade="Upgrading" if upgrade else "Installing"))
    args = [pip_name, "--verbose"]
    if upgrade and "--upgrade" not in args:
        args.append("--upgrade")
    logger.debug("Trying to install \"{pip_name}\" with pip using the following arguments: {pip_args}...".format(
        pip_name=pip_name, pip_args=args))
    return pip_install(args)


def upgrade(pip_name):
    install_only(pip_name, upgrade=True)


def find_submodules(main_package):
    packages = [main_package]
    for package in find_packages(where=main_package):
        packages.append(main_package + "." + package)
    return packages



# -*- coding: utf-8 -*-
from functools import wraps

try:
    from django.utils.decorators import available_attrs
except ImportError:  # pragma nocover
    from ..dependencies import import_or_install

    available_attrs = import_or_install("django.utils.decorators.available_attrs", "django")
# end try

from ..eastereggs.headers import get_headers  as get_easteregg_headers

__author__ = 'luckydonald'


def headers(*headers_as_dict, **headers_as_kwargs):
    """Decorator adding arbitrary HTTP headers to the response.

    This decorator adds HTTP headers specified in the argument (map), to the
    HTTPResponse returned by the function being decorated.

    Example:

    >>> @headers({'Refresh': '10', 'X-Bender': 'Bite my shiny, metal ass!'})
    >>> @headers(Refresh='10', X_Bender='Bite my shiny, metal ass!')
    >>> def index(request):
    >>> 	pass  # your stuff.

    Modified from https://djangosnippets.org/snippets/275/
    """

    def headers_wrapper(fun):
        @wraps(fun, assigned=available_attrs(fun))
        def wrapped_function(*args, **kwargs):
            response = fun(*args, **kwargs)
            for arg in headers_as_dict:
                if isinstance(arg, dict):
                    for k, v in headers_as_kwargs.items():
                        response[k] = v
                else:
                    raise TypeError("Parameter is not type dict but {}".format(type(arg)))
            for kwarg_key, kwarg_value in headers_as_kwargs.items():
                if not isinstance(kwarg_value, str):
                    # try:
                    #	temp = str(kwarg_value)
                    #	kwarg_value = temp
					# except:
                    raise TypeError("Keyword-parameter '{param}' is not type str but {type}, and str() failed.".format(
                        param=kwarg_key, type=type(kwarg_value)))
                if isinstance(kwarg_value, str):
                    response[kwarg_key.replace("_", "-")] = kwarg_value
                else:
                    raise TypeError("Keyword-parameter '{param}' is not type str but {type}.".format(param=kwarg_key,
                                                                                                     type=type(
                                                                                                         kwarg_value)))
            return response

        return wrapped_function

    return headers_wrapper


def header(name, value):
    # View decorator that sets a response header.
    #
    # Example:
    # @header('X-Powered-By', 'Django')
    # def view(request, ...):
    #     ....
    #
    # For class-based views use:
    # @method_decorator(header('X-Powered-By', 'Django'))
    # def get(self, request, ...)
    #     ...
    #
    # Taken from https://djangosnippets.org/snippets/2895/
    def decorator(func):
        @wraps(func, assigned=available_attrs(func))
        def inner(request, *args, **kwargs):
            response = func(request, *args, **kwargs)
            response[name] = value
            return response

        return inner

    return decorator


def easteregg_headers(func):
    return headers(get_easteregg_headers())(func)



# -*- coding: utf-8 -*-
from collections import defaultdict

__author__ = 'luckydonald'


def etree_to_dict(t):
    """
    Function to modify a xml.etree.ElementTree thingy to be a dict.
    Attributes will be accessible via ["@attribute"],
    and get the text (aka. content) inside via ["#text"]

    TESTED ONLY FOR PYTHON 3! (but should be working in Python 2...)
    :param t:
    :return:
    """
    # THANKS http://stackoverflow.com/a/10077069

    d = {t.tag: {} if t.attrib else None}
    children = list(t)
    if children:
        dd = defaultdict(list)
        for dc in map(etree_to_dict, children):
            for k, v in dc.items():
                dd[k].append(v)
        d = {t.tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}  # .items() is bad for python 2
    if t.attrib:
        d[t.tag].update(('@' + k, v) for k, v in t.attrib.items())  # .items() is bad for python 2
    if t.text:
        text = t.text.strip()
        if children or t.attrib:
            if text:
                d[t.tag]['#text'] = text
        else:
            d[t.tag] = text
    return d



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from luckydonaldUtils.logger import logging

__author__ = 'luckydonald'

logger = logging.getLogger(__name__)
if __name__ == '__main__':
    logging.add_colored_handler(level=logging.DEBUG)
# end if


def split_path(path):
    """
    Splits the text and build a nice import statement from it.
    Note: only well defined import paths are supported. Not something invalid like '..foo.bar..'.


    >>> split_path('foo.bar.Batz')
    ('foo.bar', 'Batz')

    >>> split_path('..lel')
    ('..', 'lel')

    >>> split_path('.lul')
    ('.', 'lul')

    >>> split_path('lol')
    ('', 'lol')

    >>> split_path('...na.na.na.na.na.na.Batman')
    ('...na.na.na.na.na.na', 'Batman')

    >>> split_path('...........yolo.swagger')
    ('...........yolo', 'swagger')

    :param path: The path to split.
    :type  path: str

    :return: The import text, like `from x import y` or `import z`
    :rtype: tuple(str)|Tuple[str, str]
    """
    last_dot_position = path.rfind('.')
    if last_dot_position == -1:
        # no dot found.
        import_path = ''
        import_name = path
    else:
        import_path = path[:last_dot_position + 1]
        import_name = path[last_dot_position + 1:]

        # handle 'foo.bar.Baz' not resulting in 'foo.bar.', i.e. remove the dot at the end.
        if import_path.rstrip('.') != '':
            # e.g. not '....'
            import_path = import_path.rstrip('.')
    # end if

    return import_path, import_name
# end def


def path_to_import_text(path):
    """
    Splits the text and build a nice import statement from it.
    Note: only well defined import paths are supported. Not something invalid like '..foo.bar.'

    >>> path_to_import_text('foo.bar.Batz')
    'from foo.bar import Batz'

    >>> path_to_import_text('..lel')
    'from .. import lel'

    >>> path_to_import_text('.lul')
    'from . import lul'

    >>> path_to_import_text('lol')
    'import lol'

    >>> path_to_import_text('...na.na.na.na.na.na.Batman')
    'from ...na.na.na.na.na.na import Batman'

    >>> path_to_import_text('...........yolo.swagger')
    'from ...........yolo import swagger'

    :param path: The path to split.
    :type  path: str

    :return: The import text, like `from x import y` or `import z`
    :rtype: str
    """
    import_path, import_name = split_path(path)

    if import_path:
        return 'from {import_path} import {import_name}'.format(import_path=import_path, import_name=import_name)
    # end if
    return 'import {import_name}'.format(import_name=import_name)
# end def



# -*- coding: utf-8 -*-
from .logger import logging
from .functions import deprecated

__author__ = 'luckydonald'

logger = logging.getLogger(__name__)

@deprecated(message="Instead of `for obj, i in iter_with_i(iterator=iterable, i_start=start)` use `for i, obj in enumerate(iterable, start=start)`.")
def iter_with_i(iterator, i_start=0):
    """
    Yields a tuple of the iterator result and an i(iterator_result, i),
    i incrementing each time.

    Example:
        >>> some_list = ["a","b"]

        >>> for iterator_result, i in iter_with_i(some_list)
        ... 	print("iterator is {}, i is {}".format(iterator_result, i))
        iterator is a, i is 0
        iterator is b, i is 1

    :param iterator: Your iterable object (lists etc.)
    :keyword i_start: Where to start counting. Default is `0`
    """
    i = i_start
    for iterator_result in iterator:
        yield (iterator_result, i)
        i += 1
    # end for
# end def


def chunks_known_length(iterable, size, length=None):
    """
    Yield successive `size`-sized chunks from `iterable`,
    if the length of `iterable is already known or easy to compute.

    https://stackoverflow.com/a/312464/3423324

    :param iterable: The object you want to split into pieces.

    :param size: The size each of the resulting pieces should have.
    :type  size: int

    :param length: Optional. Length of the `iterable`.
                   If set to `None` (default), the length get calculated automatically.
    :type  length: None | int
    """
    if length is None:
        len(iterable)
    # end if
    for i in range(0, length, size):
        yield iterable[i:i + size]
    # end for
# end for


def chunks(iterable, size):
    """
    Yield successive chunks from `iterable`, being `size` long.

    https://stackoverflow.com/a/55776536/3423324

    :param iterable: The object you want to split into pieces.
    :param size: The size each of the resulting pieces should have.
    """
    i = 0
    while True:
        sliced = iterable[i:i + size]
        if len(sliced) == 0:
            # to suppress stuff like `range(max, max)`.
            break
        # end if
        yield sliced
        if len(sliced) < size:
            # our slice is not the full length, so we must have passed the end of the iterator
            break
        # end if
        i += size  # so we start the next chunk at the right place.
    # end while
# end def



"""
Caches a result, and is able to return it. Useful in if statements.

In python it is not possible to store the result of an expression in a variable while being inside of an if:

.. code:: python

| if temp=do_something() == 42:
|   foo(temp)


And storing it before is not an option?
(you have something very resources-expensive, or changing values)

.. code:: python

| temp = do_something()
| temp2 = do_something_else()
| if temp == 42:
|   foo(temp)
| elif temp2:
|   foo2(temp2)


Somebody need to hold that result for you:

.. code:: python

| h = Holder()
| if h(do_something()) == 42:
|   foo(h())
| elif h(do_something_else()):
|   foo2(h())

That's what `Holder` is for.
"""


class Holder(object):
    value = None

    def set(self, value):
        self.value = value
        return value

    def get(self):
        return self.value

    def call(self, *args, **kwargs):
        if len(args) == 0 and len(kwargs) == 0:
            return self.get()
        else:
            return self.set(*args, **kwargs)

    def to_string(self):
        return str(self.value)

    def to_verbose_string(self):
        return repr(self.value)

    def not_empty(self):
        return True if self.value else False

    # holder = utils.Holder();
    __call__ = call  # holder("test")
    __get__ = get  # print(holder)
    __set__ = set  # holder = "test"
    __bool__ = not_empty  # if holder:
    __str__ = to_string  # user output
    __repr__ = to_verbose_string  # debug output



from collections.abc import Mapping
# noinspection PyProtectedMember
from typing import Tuple, _VT_co, Iterator


class Kwargs(Mapping):
    __FIELDS__: Tuple[str]

    def __getitem__(self, parameter: str) -> _VT_co:
        """
        Retrieve an object variable
        :param parameter: The current class's parameter to retrieve.
        :raises KeyError: Key not found.
        :return:
        """
        ...

    def __len__(self) -> int:
        ...

    def __iter__(self) -> Iterator[str]:
        """
        The keys of this element.
        """
        ...



# -*- coding: utf-8 -*-
import re

__author__ = 'luckydonald'

_REPO_NAME_REGEX = r"[a-zA-Z0-9_.-]+"
REPO_NAME_REGEX = re.compile(_REPO_NAME_REGEX)

_USER_NAME_REGEX = r"(?:[a-zA-Z0-9]|-(?!-))+[a-zA-Z0-9]"
USER_NAME_REGEX = re.compile(_USER_NAME_REGEX)

# https://regex101.com/r/eJ6cA4/
_AT_USERNAME_REGEX = (
        r"(?:\s|^|\.|,|:|\?|!|¡|¿)@(?!-)"
        r"(?P<user>" + _USER_NAME_REGEX + r")"
                                          r"(?![^\x00-\x7F])(?:$|[^a-zA-Z0-9\-_])"
)
AT_USERNAME_REGEX = re.compile(_AT_USERNAME_REGEX)

# backwards compatibility:
from .urls.github import FILE_URL_REGEX, SIMPLE_URL_REGEX



# -*- coding: utf-8 -*-
try:
    from django.http import JsonResponse, HttpResponse
    from django.conf import settings
    from django.utils.decorators import available_attrs
except ImportError:  # pragma nocover
    from ..dependencies import import_or_install

    available_attrs = import_or_install("django.utils.decorators.available_attrs", "django")
    JsonResponse = import_or_install("django.http.JsonResponse", "django")
    HttpResponse = import_or_install("django.http.HttpResponse", "django")
    settings = import_or_install("django.conf.settings", "django")
# end try
from functools import wraps
from ..logger import logging  # pip install luckydonald-utils

__author__ = 'luckydonald'

logger = logging.getLogger(__name__)


def json_response(content=None, status=None, statusText=None, exception=None):
    msg = {"status": 200, "statusText": "OK", "content": None}
    if isinstance(exception, Exception):
        logger.warn("Json Exception", exc_info=True)
        msg["status"] = 500
        msg["statusText"] = "INTERNAL SERVER ERROR"
        msg["content"] = "Generic Error"
        if settings.DEBUG:
            msg["exception"] = str(exception)
            # end if
    # end if
    if status:
        msg["status"] = int(status)
    if statusText:
        msg["statusText"] = statusText
    if content:
        msg["content"] = content
    # end if
    return JsonResponse(msg, status=msg["status"])


####
#
#  Render Exceptions
#
####

def render_all_exceptions(function):
    """
    View decorator that renders Exception in django.
    If you want to catch and respond specific Exception types, see `render_specific_exception` decorator

    >>> @render_all_exceptions
    ... def foobar()
    ... 	pass # code goes here

    """

    @wraps(function, assigned=available_attrs(function))
    def function_callable(request, *args, **kwargs):
        try:
            response = function(request, *args, **kwargs)
        except Exception as e:
            logger.debug("Exception response, because expected Exception occurred. {err_str}".format(err_str=str(e)))
            return HttpResponse(str(e), status=500)
        # end try
        return response

    return function_callable


def render_DoOutputException(function):
    """
    View decorator that renders DoOutputException's message in django.
    If you want to catch and respond specific Exception types, see `render_specific_exception` decorator.
    If you want to catch and display all Exception types, see `render_all_exceptions` decorator

    >>> @render_DoOutputException
    ... def foobar()
    ... 	raise ValueError("lol.")

    """

    @wraps(function, assigned=available_attrs(function))
    def function_callable(request, *args, **kwargs):
        try:
            response = function(request, *args, **kwargs)
        except DoOutputException as e:
            logger.debug(
                "Exception response, because expected DoOutputException occurred. {err_str}".format(err_str=str(e)))
            return HttpResponse(str(e), status=500)
        # end try
        return response

    return function_callable


def render_specific_exception(exception_class, exception_render_func=None):
    """
     View decorator that allows to render specified Exception Types.
    You can specify a `exception_render_func` which will be called with the request and the exception as arguments.

    :param   exception: The Exception you want to be catched and rendered as response.
    :type    exception: class (inherits Exception)

    :keyword exception_render_func: an optional function to render the exception. It will be called with `func(request, e)`
    :type    exception_render_func: function (2 parameters [request, exception])

    :return: Django Response
    :rtype : HttpResponse




    Example:
     @render_specific_exception(django.http.Http404)
        def view(request, ...):
            ...

    For class-based views use:
    @method_decorator(render_specific_exception(django.http.Http404))
     def get(self, request, ...)
        ...

    Import it
    >>> from luckydonaldUtils.djangos import render_specific_exception


    How to use
    >>> @render_specific_exception(AttributeError)
    ... @render_specific_exception(ValueError, exception_render_func=foo_error_renderer)  # foo_error_renderer see below.
    ... def django_foo(do_fail = 0):
    ... 	if do_fail == 1:
    ... 		raise AttributeError("ololol! Without custom renderer")
    ... 	elif do_fail == 2:
    ... 		raise ValueError("hua! With custom renderer.")
    ... 	elif do_fail == 3:
    ... 		raise AssertionError("Another test Error, not handled.")
    ... 	return HttpResponse("normal output")

    Example exception_render_func function
    >>> def foo_error_renderer(request, e):
    ... 	return HttpResponse("ERROR " + str(e) + " YEAH!") # but no status=500

    Try it:
    >>> django_foo(0)
    >>> django_foo(1)
    >>> django_foo(2)
    >>> django_foo(3)
    """

    def create_function(function):
        @wraps(function, assigned=available_attrs(function))
        def function_callable(request, *args, **kwargs):
            try:
                response = function(request, *args, **kwargs)
            except exception_class as e:
                logger.debug(
                    "Exception response, because expected Exception occurred. {err_str}".format(err_str=str(e)))
                if exception_render_func:
                    return exception_render_func(request, e)
                return HttpResponse(str(e), status=500)
            # end try
            return response

        return function_callable

    return create_function


class DoOutputException(Exception):
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import json
from datetime import datetime, timedelta
from typing import List, Dict

import requests

from luckydonaldUtils.logger import logging

from pytgbot.api_types.receivable.media import Sticker
from pytgbot.api_types.receivable.updates import Message, Update

from teleflask import TBlueprint

from ...regex.urls.telegram import ADDSTICKERS_REGEX

__author__ = 'luckydonald'


logger = logging.getLogger(__name__)
if __name__ == '__main__':
    logging.add_colored_handler(level=logging.DEBUG)
# end if

GETSTICKERS_API_KEY = os.getenv('GETSTICKERS_API_KEY', None)
GETSTICKERS_API_URL = os.getenv('GETSTICKERS_API_URL', 'https://api.getstickers.me/v1')   # so we can switch to staging there.
GETSTICKERS_ENABLED = GETSTICKERS_API_KEY and GETSTICKERS_API_URL

TIMEOUT = 0.5


# noinspection PyCompatibility
class BotMessage(object):
    bot_id: int
    message: Message

    def __init__(self, bot_id, message):
        self.bot_id = bot_id
        self.message = message
    # end def

    def to_array(self):
        return {"bot_id": self.bot_id, "message": self.message.to_array()}
    # end def

    def __repr__(self):
        return f'{self.__class__.__name__}(bot_id={self.bot_id!r}, message={self.message!r})'
    # end def
# end class


# noinspection PyCompatibility
class CacheData(object):
    enabled: bool
    next_send: datetime
    queue: List[BotMessage]

    def __init__(self, queue=None, *, next_send=None, enabled=True):
        if queue is None:
            queue = []
        # end if
        if next_send is None:
            next_send = datetime.now() + SEND_INTERVAL
        # end if
        self.enabled = enabled
        self.queue = queue
        self.next_send = next_send
    # end def

    def __repr__(self):
        return f'{self.__class__.__name__}(enabled={self.enabled!r}, queue={self.queue!r}, next_send={self.next_send!r})'
    # end def
# end class



SEND_INTERVAL = timedelta(minutes=1)

# noinspection PyCompatibility
SEND_CACHE_STICKER = CacheData(enabled=True)

sticker_crawl_tbp = TBlueprint(__name__)

if not GETSTICKERS_ENABLED:
    logger.warning('Did not register sticker crawler as GETSTICKERS_API_KEY is not defined.')
    raise ImportError('Did not register sticker crawler as GETSTICKERS_API_KEY is not defined.')
# end if


@sticker_crawl_tbp.on_message
def on_sticker_message(update: Update, msg: Message):
    submit_sticker_message(msg)
    collect_pack_text(msg)
    return
# end def


@sticker_crawl_tbp.on_update('channel_post')
def on_sticker_update_channel_post(update: Update):
    submit_sticker_message(update.channel_post)
    collect_pack_text(update.channel_post)
    return
# end def


@sticker_crawl_tbp.on_update('edited_message')
def on_sticker_update_edited_message(update: Update):
    submit_sticker_message(update.edited_message)
    collect_pack_text(update.edited_message)
    return
# end def


@sticker_crawl_tbp.on_update('edited_channel_post')
def on_sticker_update_edited_channel_post(update: Update):
    submit_sticker_message(update.edited_channel_post)
    collect_pack_text(update.edited_channel_post)
    return
# end def


def collect_pack_text(message: Message):
    if not message.text or '/addstickers/' not in message.text:
        return
    # end if
    for match in ADDSTICKERS_REGEX.finditer(message.text):
        submit_pack(match.group('pack'))
    # end for
# end def


def submit_pack(pack_url: str):
    try:
        requests.put(
            GETSTICKERS_API_URL + '/submit/pack',
            params={
                "key": GETSTICKERS_API_KEY,
                "bot_id": sticker_crawl_tbp.user_id,
                "pack": pack_url,
            },
            timeout=TIMEOUT,
        )
    except requests.HTTPError as e:
        try:
            result = repr(e.response.json())
        except:
            result = e.response.text
        # end try
        logger.warning(f'Submitting sticker to getstickers.me failed with error code {e.response.status_code}: {result}')
    except:
        logger.warning('Submitting sticker to getstickers.me failed.', exc_info=True)
    # end try
# end def


# noinspection PyCompatibility
def submit_sticker_message(message: Message):
    if not isinstance(message, Message) or not isinstance(message.sticker, Sticker):
        return
    # end if
    global SEND_CACHE_STICKER
    wrapped_msg = BotMessage(bot_id=sticker_crawl_tbp.user_id, message=message)
    if SEND_CACHE_STICKER.enabled:
        SEND_CACHE_STICKER.queue.append(wrapped_msg)

        if datetime.now() < SEND_CACHE_STICKER.next_send:
            logger.debug(f'Not yet sending the sticker, waiting until {SEND_CACHE_STICKER.next_send}.')
            return
        # end if
        SEND_CACHE_STICKER.next_send = datetime.now() + SEND_INTERVAL
        messages_to_send, SEND_CACHE_STICKER.queue = SEND_CACHE_STICKER.queue, []
    else:
        messages_to_send = [wrapped_msg]
    # end if
    if not messages_to_send:
        logger.debug(f'no stickers, not sending.')
        return
    # end if
    messages_to_send = [msg.to_array() for msg in messages_to_send]
    logger.debug(f'sending {messages_to_send!r} to the API with key={GETSTICKERS_API_KEY!r}.')
    try:
        result = requests.put(
            GETSTICKERS_API_URL + '/submit/stickers',
            params={
                "key": GETSTICKERS_API_KEY,
            },
            json=messages_to_send,
            timeout=TIMEOUT,

        )
        result.raise_for_status()
    except requests.HTTPError as e:
        try:
            result = repr(e.response.json())
        except:
            result = e.response.text
        # end try
        logger.warning(f'Submitting stickers to getstickers.me failed with error code {e.response.status_code}: {result}')
    except:
        logger.warning('Submitting stickers to getstickers.me failed.', exc_info=True)
    # end try
# end def



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from luckydonaldUtils.logger import logging

__author__ = 'luckydonald'

logger = logging.getLogger(__name__)
if __name__ == '__main__':
    logging.add_colored_handler(level=logging.DEBUG)
# end if

"""
package/            # package
    __init__.py     # package
    subpackage1/    # package.subpackage1
        __init__.py # package.subpackage1
        moduleX.py  # package.subpackage1.moduleX
        moduleY.py  # package.subpackage1.moduleY
    subpackage2/    # package.subpackage2
        __init__.py # package.subpackage2
        moduleZ.py  # package.subpackage2.moduleZ
    moduleA.py      # moduleA

Assuming that the current file is either moduleX.py or subpackage1/__init__.py, following are correct usages of the new syntax:
package.subpackage1.moduleX


# from .moduleY import spam
assert relimport('package.subpackage1.moduleY.spam', 'package.subpackage1.moduleX') == '.moduleY.spam'

# from . import moduleY
assert relimport('package.subpackage1.moduleY', 'package.subpackage1.moduleX') == '.moduleY'

# from ..subpackage1 import moduleY
assert relimport('package.subpackage1.moduleY', 'package.subpackage1.moduleX') == '..subpackage1.moduleY'

# from ..subpackage2.moduleZ import eggs
assert relimport('package.subpackage2.moduleZ', 'package.subpackage1.moduleX') == '..subpackage2.moduleZ.eggs'

# from ..moduleA import foo
assert relimport('moduleA.foo', 'package.subpackage1.moduleX') == '..moduleA.foo'

# from ...package import bar
assert relimport('package.bar', 'package.subpackage1.moduleX') == '..package.bar'

# from ...sys import path
assert relimport('sys.path', 'package.subpackage1.moduleX') == '..sys.path'

assert relimport('foo.bar.batz', 'foo.bar') == '.batz'

"""


def relimport(destination, start):
    """
    Return a relative import path from the `start` module's view'.
    This is a path computation: the filesystem is not accessed to confirm the existence or nature of `destination` or `start`.

    Here are some examples:

        >>> relimport(destination='foo.bar.batz', start='foo.bar')  # in foo.bar: from . import batz: .batz
        '.batz'

        >>> relimport(destination='foo.bar.batz.gneeh', start='foo.bar')
        '.batz.gneeh'

        >>> relimport(destination='foo.bar', start='foo.bar.batz')  # in foo.bar.batz: from ... import batz
        '...bar'

        >>> relimport(destination='foo.bar', start='foo.bar.batz.gneeh')  # in foo.bar: from . import batz: .batz
        '....bar'

        >>> relimport(destination='foo.bar.c', start='foo.bar.batz')
        '.c'

        >>> relimport(destination='foo.b.c', start='foo.bar.batz')
        '..b.c'

        >>> relimport(destination='package.subpackage1.moduleY.spam', start='package.subpackage1.moduleX')
        '.moduleY.spam'

        >>> relimport(destination='package.subpackage1.moduleY', start='package.subpackage1.moduleX')
        '.moduleY'

        >>> relimport(destination='package.subpackage2.moduleZ.eggs', start='package.subpackage1.moduleX')
        '..subpackage2.moduleZ.eggs'

        >>> relimport(destination='moduleA.foo', start='package.subpackage1.moduleX')
        '...moduleA.foo'

        >>> relimport(destination='sys.path', start='package.subpackage1.moduleX')
        '...sys.path'

        >>> relimport(destination='foo.bar.batz', start='foo.bar')
        '.batz'

        >>> relimport(destination='foo.bar.batz.Something', start='foo.bar.gnerf')
        '.batz.Something'

    Note: Those examples above are also used as part of the unit tests, via the doctest system.
    Note: This is called the same way as `os.path.relpath(path, start)` would be, and is compatible in calling,
          except the `path` parameter called `destination` to be more clear, and `start` (not yet) being optional.

    :param destination: the destination path
    :type  destination: str

    :param start: our current module where we call the import. You can supply the current module with `start=__NAME__`.
    :type  start: str

    :return: the relative path. `'...foo.bar'`
    :rtype: str
    """
    destination_parts = destination.split('.')
    start_parts = start.split('.')
    original_destination_parts = destination.split('.')

    for i in range(min(len(start_parts), len(destination_parts))):
        if start_parts[0] == destination_parts[0]:
            start_parts.pop(0)
            destination_parts.pop(0)
            continue
        else:
            break
        # end if
    # end for
    if destination_parts == [] and start_parts == []:
        # same path
        # a.b.c == a.b.c
        return None  # same dir
    # end if
    if start_parts == [] and destination_parts:
        # some children module, even deeper in the tree
        # a == a.b.c
        return "." + ".".join(destination_parts)
    # end if
    if start_parts and destination_parts == []:
        # some parent folder module
        # a.b.c == a
        # make sure it's ending with a package name to import.
        return '.' + "." * len(start_parts) + '.' + original_destination_parts[-1]
    # end if
    if start_parts and destination_parts:
        # so the paths differ at some point, or from the beginning.
        # a.b.c == x.y.z
        return '.' + "."*(len(start_parts)-1) + ".".join(destination_parts)
    # end if
# end def



try:
    from collections.abc import Mapping  # py3
except ImportError:
    from collections import Mapping  # py2
# end try


class Singleton(type):
    """
        Only one Instance of the class.
        Like in Java .getInstance()


        Make a class Singleton:

        Python 3:
        apply with keyword argument in the base-class list:
        >>> class Example(object, metaclass = Singleton):
        >>>    pass

        Python 2.7:
        apply like
        >>> class Exampe(object):
        >>>    __metaclass__ = Singleton
        >>>    pass


        Usage of the class:

        At the first call you have to supply the parameters for class, as usual.
        Afterwards you can omit them, they will be ignored if given.

        Create the instance:
        >>> whatever = Example("arguments", "go", "here", as_well_as="keyword args")

        Use a already created instance:
        >>> something = Example()  # note: No args or kwargs required here.
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            args = args if args else []
            kwargs = kwargs if kwargs else {}
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]
    # end def
# end class


class Kwargs(Mapping):
    __FIELDS__ = tuple()

    def __getitem__(self, parameter):
        """
        Retrieve an object variable
        :param parameter: The current class's parameter to retrieve.
        :raises KeyError: Key not found.
        :return:
        """
        if parameter in self.__FIELDS__:
            return getattr(self, parameter)
        # end if
        raise KeyError("Key {!r} not found".format(parameter))
    # end def

    def __len__(self):
        return len(self.__FIELDS__)
    # end def

    def __iter__(self):
        return iter(self.__FIELDS__)
    # end def

    def __repr__(self):
        return "{name}({args})".format(
            name=self.__class__.__name__,
            args=", ".join("{k}={v!r}".format(k=k, v=self[k]) for k in self.__FIELDS__)
        )
    # end def
# end class




# -*- coding: utf-8 -*-
# noinspection
__author__ = 'luckydonald'

def get_local_time(time:datetime.datetime, time_zone:str=None) -> datetime.datetime:
	if time_zone is None:
		time_zone = "Europe/Berlin" # todo: config
	return timezone('UTC').localize(time).astimezone(timezone(time_zone))


def loop_seconds_between_times(datetime1:datetime.datetime, datetime2:datetime.datetime):
	delta = datetime2 - datetime1
	seconds = delta.total_seconds()
	for x in range(0, int(seconds)):
		yield (datetime1 + datetime.timedelta(seconds=x))
	yield datetime2  # so loop_seconds_between_times(dat,dat) is at least called once with dat.


# -*- coding: utf-8 -*-
import re

__author__ = 'luckydonald'
__all__ = [
    'TELEGRAM_DOMAIN_REGEX', '_TELEGRAM_DOMAIN_REGEX',
    'USER_LINK_REGEX', '_USER_LINK_REGEX',
    'ADDSTICKERS_REGEX', '_ADDSTICKERS_REGEX',
]

_TELEGRAM_DOMAIN_REGEX = r'(?:https?://)?(?:t(?:lgrm)?\.me|telegram\.(?:me|dog))'
TELEGRAM_DOMAIN_REGEX = re.compile(_TELEGRAM_DOMAIN_REGEX)

from ..telegram import _USERNAME_REGEX

_USER_LINK_REGEX = '(?P<domain>' + _TELEGRAM_DOMAIN_REGEX + ')/(?P<username>' + _USERNAME_REGEX + ')'
USER_LINK_REGEX = re.compile(_USER_LINK_REGEX)

_ADDSTICKERS_REGEX = '(?P<domain>' + _TELEGRAM_DOMAIN_REGEX + r")/addstickers/(?P<pack>\w{5,})\b"  # regex101.com/r/MvQWDv/1
ADDSTICKERS_REGEX = re.compile(_ADDSTICKERS_REGEX)



from typing import Union, List, Tuple

WAIFU = [
    "Littlepip",
    "Toaster Repair Pony",
    "Littlepip",
    "The Lightbringer",
    "The Stable Dweller"
]
BEST_PONY = [
    "Applejack",
    "Daring Do",
    "Littlepip",
]


""":var UNWORTHY: A list of either strings, or a fun-fact tuple."""
UNWORTHY: List[Union[str, Tuple[str, str, str]]] = [
    "Blackjack",
    "Blackjerk",
    "oc:blackjack",
    "Security Mare (Blackjack)",
    "Maiden of the Stars (Blackjack)",
    "Queen Whiskey (Blackjack)",
    "Queen Blackjack",
    "Black-And-Red OC",
    "Blackjack (Cyber-Unicorn with Cyber Wings)",
    "Blackjack (Cyber Alicorn)",
    "Blackjack (Unicorn Blank/Clone again after her original body was torn in two)",
    "Blackjack (Biological Great-great-great-great-great-great-great-great-great Granddaughter of Twilight Sparkle)",
    "Blackjack (Biological Great-great-great-great-great-great-great-great-great Granddaughter of Big Macintosh)",
    "Blackjack Clone (Made out of Discords blood)",
    ("Blackjack", "Did you know:", "Blackjack died up to 4 times:\n\nIntroduction (Alive): Chp. 1\nDead: Chp. 33\nAlive again: Chp. 34\nDead: Chp. 48\nAlive: Chp. 49\nPossibly Dead: Chp. 54\nAlive Chp.54 \nDead end of Chp.77 (Original body)\nAlive in Epilogue (Clone body, implied to be immune to age)"),
    ("Blackjack", "Did you know:", "Blackjack has had two boats dropped on her in her time in Hoofington - three if you count an airship."),
    ("Blackjack", "Did you know:", "In an interview with the fan website Equestria Daily, Somber was originally going to have Blackjack be killed off by the end of Chapter 33 while Rampage and Scotch Tape continue where she left off."),
    ("Blackjack", "Did you know:", "Another running gag is Blackjack's ignorance that Littlepip is the famed Stable Dweller."),
]



try:
    from gitinfo_values import __all__
except:
    __all__ = ["GIT_COMMIT", "GIT_COMMIT_SHORT", "GIT_DIRTY_PROJECT", "GIT_DIRTY_GLOBAL", "GIT_MESSAGE",
               "GIT_AUTHOR", "GIT_DATE", "GIT_HEAD", "GIT_BRANCHES", "GIT_TAGS", "VERSION_STR",
               "GIT_MODIFIED_FILES", "GIT_MODIFIED_FILES_STR"]
# end try

try:
    from gitinfo_values import GIT_COMMIT
except:
    GIT_COMMIT = None
# end try

try:
    from gitinfo_values import GIT_DIRTY_PROJECT
except:
    GIT_DIRTY_PROJECT = None
# end try

try:
    from gitinfo_values import GIT_DIRTY_GLOBAL
except:
    GIT_DIRTY_GLOBAL = None
# end try

try:
    from gitinfo_values import GIT_COMMIT_SHORT
except:
    GIT_COMMIT_SHORT = None
# end try

try:
    from gitinfo_values import GIT_MESSAGE
except:
    GIT_MESSAGE = None
# end try

try:
    from gitinfo_values import GIT_AUTHOR
except:
    GIT_AUTHOR = None
# end try

try:
    from gitinfo_values import GIT_AUTHOR
except:
    GIT_AUTHOR = None
# end try

try:
    from gitinfo_values import GIT_DATE
except:
    GIT_DATE = None
# end try

try:
    from gitinfo_values import GIT_HEAD
except:
    GIT_HEAD = None
# end try

try:
    from gitinfo_values import GIT_BRANCHES
except:
    GIT_BRANCHES = []
# end try

try:
    from gitinfo_values import GIT_TAGS
except:
    GIT_TAGS = []
# end try

try:
    from gitinfo_values import VERSION_STR
except:
    VERSION_STR = "unknown"
# end try

try:
    from gitinfo_values import GIT_MODIFIED_FILES
except:
    GIT_MODIFIED_FILES = {"added": [], "copied": [], "deleted": [], "modified": [], "renamed": [], "type_changed": [],
                          "unmerged": [], "unknown": [], "broken": [], "by_file": {}}
# end try

try:
    from gitinfo_values import GIT_MODIFIED_FILES_STR
except:
    GIT_MODIFIED_FILES_STR = ""
# end try

try:
    from teleflask.messages import HTMLMessage
    from teleflask import TBlueprint

    version_tbp = TBlueprint(__name__)


    @version_tbp.command('version')
    def cmd_version(update, text):
        return HTMLMessage('<code>{version}</code>'.format(version=escape(VERSION_STR)))
    # end def
except ImportError:
    pass
# end try

try:
    from html import escape
    from flask import Blueprint

    version_bp = Blueprint('version', __name__)


    @version_bp.route('/version/')
    def route_version():
        return VERSION_STR
    # end def
except ImportError:
    pass
# end try



# -*- coding: utf-8 -*-
from .. import py3

__author__ = 'luckydonald'


if py3:
    def binary_ip_to_str(host):
        return ".".join([str(x) for x in host])
else:
    def binary_ip_to_str(host):
        return ".".join([str(ord(x)) for x in host])


from datetime import datetime, timedelta

import functools
import inspect
import logging

from .clazzes import Kwargs

logger = logging.getLogger(__name__)

__author__ = 'luckydonald'
__all__ = [
    "caller", "deprecated", "gone", "cached",
    "CallerResult", "CallerInfo"
]


class CallerResult(Kwargs):
    """
    Contains the data of who called us.

    Can also be used via ** mapping.
    >>> x = CallerResult("aaa", "bbb")
    >>> dict(**x)
    {'self': 'aaa', 'caller': 'bbb'}
    """
    __FIELDS__ = ("self", "caller")

    # noinspection PyMethodParameters
    def __init__(this, self, caller):
        """
        :type  self: CallerInfo
        :type  caller: CallerInfo
        """
        this.self = self
        this.caller = caller
    # end def
# end class


class CallerInfo(Kwargs):
    __FIELDS__ = ("name", "file", "line", "code")

    def __init__(self, name, file, line, code=None):
        self.name = name
        self.file = file
        self.line = line
        self.code = code
    # end def
# end class


def caller(level=0, kwarg_name="call"):
    """
    functions decorated with this will be called with an `call` kwarg, containing information about the function itself, and the caller.
    If the caller could not be fetched correctly, the `caller`s attributes all will be `None`.

    Level is the amount of functions to go upwards. Default is 1.
    kwarg_name is the name of the kwarg parameter we input the caller data as. Default is "call".


    Usage:

    >>> @caller
    ... def whatever(arg, kwarg=None, call=None):
    ...     pass  # do something with the `call` info

    >>> @caller(1)  # change this if you you are deep within another call. You can adust the stack index here.
    ... def whatever(arg, kwarg=None, call=None):
    ...     pass  # do something with the `call` info

    >>> @caller(kwarg_name="_call_info")  # change this if you you are deep within another call. You can adust the stack index here.
    ... def whatever(arg, kwarg=None, _call_info=None):
    ...     pass  # do something with the `_call_info` info
    """
    def caller_func_wrapper(func, level=0, kwarg_name="call"):
        """
        :param func: The function we need to call.

        :param level: The level to go back in stack.
        :type  level: int

        :param kwarg_name: the **kwargs element to write to.
        :type  kwarg_name: str
        :return:
        """

        @functools.wraps(func)
        def caller_wrapping_func(*args, **kwargs):
            # so the decorated function is the 'self' one.
            try:
                self = CallerInfo(
                    name=func.__name__,
                    file=func.__code__.co_filename,
                    line=func.__code__.co_firstlineno + 1,
                )
            except AttributeError as e:
                    logger.debug("self lookup failed", exc_info=True)
                    # we just don't know
                    self = CallerInfo(
                        name=None,
                        file=None,
                        line=None,
                    )
                # end if
            # end try
            try:
                s = inspect.stack()
                # +1 as we need to ignore the caller_wrapping_func
                stack = s[level + 1]  # the +1 out-of-bounds is handled by the except IndexError.
                caller = CallerInfo(
                    name=stack[3],
                    file=stack[1],
                    line=stack[2],
                    code=stack[4][0],
                )
            except (TypeError, IndexError):
                caller = CallerInfo(
                    name=None,
                    file=None,
                    line=None,
                    code=None,
                )
            # end try
            call = CallerResult(
                self=self,
                caller=caller
            )
            kwargs[kwarg_name] = call
            return func(*args, **kwargs)

        return caller_wrapping_func
    # end def

    def caller_wrapping_func_applyer(func):
        return caller_func_wrapper(func, level=level, kwarg_name=kwarg_name)
    # end def

    try:
        is_callable = callable(level)
    except (SyntaxError, NameError):
        is_callable = inspect.isfunction(level) or inspect.isbuiltin(level)
    # end try
    if is_callable:  # level was not provided, level is actually our function to decorate
        func = level
        _func = functools.wraps(func)(caller_func_wrapper)
        return _func(func)  # keeping the defaults at the defaults.
    return caller_wrapping_func_applyer  # else: level is level, we need to return a decorator accepting the function
# end def


def deprecated(func, message=None):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being logged
    when the function is used."""

    if isinstance(func, str):  # @deprecated("message")
        message = func

        def inners(func):  # @deprecated("message")(func)
            return deprecated(func, message)  # @deprecated("message")(func) ->  @deprecated(func, "message")

        return inners

    # end if
    # -> @deprecated(func)

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        try:
            try:  # python 2 + python 3
                f_name = func.__name__
                f_file = func.__code__.co_filename
                f_line = func.__code__.co_firstlineno + 1
            except AttributeError:  # maybe some old python 2?
                f_name = func.func_code.co_name
                f_file = func.func_code.co_filename
                f_line = func.func_code.co_firstlineno + 1
            # end try
            try:
                logger.warning(
                    "Call to deprecated function {f_name} at file {f_file}:{f_line}\n"
                    "{msg}Called from function {c_name} at file {c_file}:{c_line}\nLine: {c_code}".format(
                        f_name=f_name, f_file=f_file,
                        f_line=f_line,
                        msg="Warning: " + message.strip() + "\n" if message else "",
                        c_name=inspect.stack()[1][3], c_file=inspect.stack()[1][1],
                        c_line=inspect.stack()[1][2], c_code=inspect.stack()[1][4][0],
                    ))
            except TypeError:
                logger.warning(
                    "Call to deprecated function {f_name} at file {f_file}:{f_line}{msg}".format(
                        f_name=f_name, f_file=f_file,
                        f_line=f_line, msg="\nWarning: " + message.strip() if message else ""
                    ))
        except (AttributeError, TypeError):
            logger.warning(
                "Call to deprecated function.{msg}".format(msg="\nWarning: " + message.strip() if message else ""),
                exc_info=True
            )
        return func(*args, **kwargs)

    return new_func  # @deprecated(func)


# end def
"""
Test would be
>> @deprecated
>> def foo(): pass

>> @deprecated("Custom Message")
>> def foo(): pass
"""


def gone(func):
    """This is a decorator which can be used to mark functions
    as deprecated, and gone. It will raise a NotImplementedError
    when the function is used."""
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        raise NotImplementedError(
            "Call to deprecated function {} at file {}:{}. Called from function {} at file {}:{}\nLine: {}".format(
                func.__name__, func.__code__.co_filename,  # python2: func.func_code.co_filename
                func.__code__.co_firstlineno + 1,  # python2: func.func_code.co_firstlineno
                inspect.stack()[1][3], inspect.stack()[1][1], inspect.stack()[1][2], inspect.stack()[1][4][0]
            )
        )

    # end def
    return new_func
# end def


def cached(max_age=None):
    """
    Cache a function.
    You can specify a max_age (:class:`datetime.timedelta`) after when the function will be called again.

    :keyword max_age: timedelta of how long we should keep the cache. None means forever.
    :type    max_age: None or timedelta
    """
    # http://stackoverflow.com/a/30698822/3423324
    # _format = {"returns": "I am a return value!!", "max_age": timedelta(seconds=60), "last_hit": datetime.now()}
    memo = {}
    import json

    def stringify(obj):
        if isinstance(obj, (list, tuple)):
            return repr([stringify(x) for x in obj])
        if isinstance(obj, dict):
            res = {}
            for k, v in obj.items():
                res[k] = stringify(k)
            # end for
            return repr(res)
        try:
            return json.dumps(obj)
        except TypeError:
            return str(repr(obj)) + str(obj) + str(id(obj))
            # end if

    def func_wrapper(function, **decorator_kwargs):
        def wrapper(*args, **kwargs):
            args_key = stringify([args, kwargs])
            now = datetime.now()
            if "max_age" in decorator_kwargs:
                max_age_ = decorator_kwargs["max_age"]
            else:
                max_age_ = max_age
            if (args_key) in memo.keys():
                logger.debug(memo[args_key])
                if not max_age_ or now - memo[args_key]["last_hit"] < memo[args_key]["max_age"]:
                    return memo[args_key]["returns"]
                    # end if
            # end if
            res = function(*args, **kwargs)
            memo[args_key] = {"returns": res, "last_hit": now, "max_age": max_age_}
            return res

        return wrapper

    try:
        is_callable = callable(max_age)
    except (SyntaxError, NameError):
        is_callable = inspect.isfunction(max_age) or inspect.isbuiltin(max_age)
    # end try
    if is_callable:  # max_age was not provided, max_age is actually our function to decorate
        func = max_age
        return func_wrapper(func, max_age=None)
    return func_wrapper  # else: max_age is max_age, we need to return a decorator accepting the function
# end def



# -*- coding: utf-8 -*-


from ..dependencies import import_or_install

try:
    from PIL import Image
except ImportError:  # pragma nocover
    Image = import_or_install("PIL.Image", "Pillow")  # pip install Pillow
# end try

__author__ = 'luckydonald'


def most_frequent_color(image, colors=10):
    image2 = image.convert("P", palette=Image.ADAPTIVE, colors=colors)
    image3 = image2.convert(image.mode)
    del image2

    w, h = image3.size
    pixels = image3.getcolors(w * h)

    most_frequent_pixels = [pixels[0]]

    for count, color in pixels:
        length = len(most_frequent_pixels)
        for i in range(0, length):
            if color == most_frequent_pixels[i][1]:
                break
            if count > most_frequent_pixels[i][0]:
                most_frequent_pixels.insert(i, (count, color))
                break
            elif count < most_frequent_pixels[length - 1]:
                most_frequent_pixels.append((count, color))
                break
    del image3
    return most_frequent_pixels[:colors]



# -*- coding: utf-8 -*-
from luckydonaldUtils.logger import logging  # pip install luckydonald-utils
from .functions import __all__ as _functions_all
# noinspection PyUnresolvedReferences
from .functions import *  # because everyone searches for the decorators in this file ... lel.

__author__ = 'luckydonald'
logger = logging.getLogger(__name__)

# noinspection PyProtectedMember

__all__ = [
    "decorator_with_default_params", "classproperty",
] + _functions_all

# noinspection PyProtectedMember


# TODO: py2 safe?


def decorator_with_default_params(real_decorator, args, kwargs, default_args=None, default_kwargs=None):
    """
    This function makes it easy to build a parametrized decorator, having a default value.
    Construct your decorator like this:

    >>> def decorator(*d_args, **d_kwargs): # d_args with d like decorator
    ... 	def function_processor(function, *args, **kwargs):
    ... 		print("additional_args", args, kwargs)
    ... 		# just do stuff here with the function, or with the args.
    ... 		return function
    ... 	return decorator_with_default_params(function_processor, d_args, d_kwargs, default_args=["I am a default argument, used if you don't provide args"])

    so you can use that decorator:

    >>> @decorator # will fall back to the default value.
    ... @decorator("hey") # with parameter(s).
    ... @decorator(key="value") # with parameter(s).
    ... def foobar()
    ... 	pass
    ... foobar = decorator(foobar) # use default value # same as @decorator
    ... foobar = decorator(foobar, "hey") # with parameter(s) # same as @decorator("hey")
    ... foobar = decorator(foobar, key="value") # with parameter(s) # same as @decorator(key="value")

    :param real_decorator: The function whitch gets a function, and *args. Must return the function again.
    :param default_args_list: arguments which will be provided to the real_decorator, if @decorator is used without params
    :param args:
    :param kwargs:
    :return:
    """
    if default_args is None:
        default_args = []
    if default_kwargs is None:
        default_kwargs = {}
    if args is not None and len(args) > 0 and callable(args[0]):
        func = args[0]
        if len(args) == 1:  # @admin  or  func = admin(func)
            if kwargs is None or len(kwargs) == 0: # no arguments/kwargs
                return real_decorator(func, *default_args, **default_kwargs)  # func = arg[0]
            else:  # has kwargs
                return real_decorator(*args, **kwargs) # args contains func.
            # end if
        else:  # test = admin(test, param)
            return real_decorator(*args, **kwargs)
        # end if
    else:  # @admin(param)
        if args is None or len(args) == 0: # you don't provide a function at all.
            if kwargs is None or len(kwargs) == 0:
                args = default_args
                kwargs = default_kwargs
            # end if
        # end if
        elif args[0] == Ellipsis: # so you can insert callables, after Ellipsis as first argument.
            args = args[1:]
        # end if

        def real_decorator_param_provider(func):
            return real_decorator(func, *args, **kwargs)
        # end def

        return real_decorator_param_provider  # it will call real_decorator(func) by its own
    # end if
# end def


# noinspection PyPep8Naming
class classproperty(property):
    # https://stackoverflow.com/a/7864317/3423324#python-staticmethod-with-property
    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()
    # end def
# end class




# -*- coding: utf-8 -*-
from typing import Union, Any, List, Dict

# from luckydonaldUtils.logger import logging

__author__ = 'luckydonald'
__all__ = ["JSONType"]

# logger = logging.getLogger(__name__)


# define a type for the returned stuff of parsed JSON,
# at least until the fine folks at https://github.com/python/typing/issues/182 find a solution.
JSONType = Union[None, bool, int, float, str, List[Any], Dict[str, Any]]
# JSONType = Union[None, bool, int, float, str, List['JSONType'], Dict[str, 'JSONType']]  # someday recursive might work





from base64 import urlsafe_b64encode, urlsafe_b64decode
from luckydonaldUtils.encoding import to_native as n, to_binary as b
import re

REPLACEMENTS = {"-": "_0", "_": "_1"}


def revert_replacements(replacements):
    return {v: k for k, v in replacements.items()}
# end def


def multi_replace(string, replacements):
    """
    Given a string and a replacement map, it returns the replaced string.
    :param str string: string to execute replacements on
    :param dict replacements: replacement dictionary {value to find: value to replace}
    :rtype: str
    """
    # Place longer ones first to keep shorter substrings from matching where the longer ones should take place
    # For instance given the replacements {'ab': 'AB', 'abc': 'ABC'} against the string 'hey abc', it should produce
    # 'hey ABC' and not 'hey ABc'
    substrs = sorted(replacements, key=len, reverse=True)

    # Create a big OR regex that matches any of the substrings to replace
    regexp = re.compile('|'.join(map(re.escape, substrs)))

    # For each match, look up the new string in the replacements
    return regexp.sub(lambda match: replacements[match.group(0)], string)


def short_custom_base64_url_decode(base, encode_replacements=REPLACEMENTS):
    replacements = revert_replacements(encode_replacements)
    base = multi_replace(n(base), replacements)
    # add missing padding # http://stackoverflow.com/a/9807138
    return n(urlsafe_b64decode(base + '='*(4 - len(base)%4)))
# end def


def short_custom_base64_url_encode(string, encode_replacements=REPLACEMENTS):
    base = n(urlsafe_b64encode(b(string))).rstrip("=")
    return multi_replace(base, encode_replacements)
# end def




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from luckydonaldUtils.logger import logging
from teleflask import TBlueprint
from .sticker import sticker_crawl_tbp

__author__ = 'luckydonald'

logger = logging.getLogger(__name__)
if __name__ == '__main__':
    logging.add_colored_handler(level=logging.DEBUG)
# end if


crawl_tbp = TBlueprint(__name__)

crawl_tbp.register_tblueprint(sticker_crawl_tbp)



# -*- coding: utf-8 -*-
import unittest
from collections import OrderedDict

from luckydonaldUtils.regex.github import FILE_URL_REGEX, AT_USERNAME_REGEX
__author__ = 'luckydonald'


class Test(unittest.TestCase):
    """
    Tests if the Regex matches only the right urls.
    """
    # input url and permanent url of that. None, if not supposed to match.
    # Lol, the permanent URLs can/will change. Should setup a extra repo for testing this.
    urls_to_test = OrderedDict()
    urls_to_test["https://github.com/luckydonald/pytg/issues/72"] = None
    urls_to_test["https://github.com/luckydonald/pytg"] = "https://github.com/luckydonald/pytg/tree/ed8d622a06fd24ec75c0c301984fa6d58ebd6e61"
    urls_to_test["https://github.com/luckydonald/pytg/"] = "https://github.com/luckydonald/pytg/tree/ed8d622a06fd24ec75c0c301984fa6d58ebd6e61"
    urls_to_test["https://github.com/luckydonald/pytg/blob/master/.gitmodules"] = "https://github.com/luckydonald/pytg/blob/ed8d622a06fd24ec75c0c301984fa6d58ebd6e61/.gitmodules"
    urls_to_test["https://github.com/luckydonald/pytg/tree/development"] = "https://github.com/luckydonald/pytg/tree/4e68eb7a13f19efe74994dcf54b9ef01de8bae3b"
    urls_to_test["https://github.com/luckydonald/pytg/tree/issue%233"] = "https://github.com/luckydonald/pytg/tree/35293c98be83672d9384c751cff7e40d7590e158"
    urls_to_test["https://github.com/luckydonald/pytg/blob/issue%233/.gitignore"] = "https://github.com/luckydonald/pytg/blob/35293c98be83672d9384c751cff7e40d7590e158/.gitignore"
    urls_to_test["https://github.com/luckydonald/pytg/blob/35293c98be83672d9384c751cff7e40d7590e158/.gitignore"] = "https://github.com/luckydonald/pytg/blob/35293c98be83672d9384c751cff7e40d7590e158/.gitignore"
    urls_to_test["https://github.com/notifications"] = None
    urls_to_test["https://github.com/luckydonald/pytg/commit/8c62dc4a35ef4ff2b81f4c0801f5448374fccaa1"] = None
    urls_to_test["https://github.com/luckydonald/pytg/pull/61/commits"] = None
    urls_to_test["https://github.com/huiyiqun/pytg/blob/e3790d9c2188949991295580f4f06d6f228ad082/pytg/fix_msg_array.py"] = "https://github.com/huiyiqun/pytg/blob/e3790d9c2188949991295580f4f06d6f228ad082/pytg/fix_msg_array.py"

    user_to_test = OrderedDict()
    user_to_test["@luckydonald"] = "luckydonald"
    user_to_test["luckydonald"] = None  # no @
    user_to_test["@with_at"] = None     # _ not allowed.
    user_to_test["without_at"] = None   # no @
    user_to_test["@bonbot"] = "bonbot"
    user_to_test["@sweetiebot"] = "sweetiebot"
    user_to_test["@mlfwbot"] = "mlfwbot"
    user_to_test["@mylittlefacewhenbot"] = "mylittlefacewhenbot"
    user_to_test["@example123"] = "example123"
    user_to_test["@look_i_have_a_username"] = None
    user_to_test["@and_more"] = None
    user_to_test["@bot_username_bot"] = None
    user_to_test["@pollbot"] = "pollbot"
    user_to_test["@gif"] = "gif"
    user_to_test["@vid"] = "vid"
    user_to_test["@pic"] = "pic"
    user_to_test["@bing"] = "bing"
    user_to_test["@wiki"] = "wiki"
    user_to_test["@imdb"] = "imdb"
    user_to_test["@bold"] = "bold"
    user_to_test["@_illegal"] = None
    user_to_test["@not__allowed"] = None
    user_to_test["@illegal_2_"] = None
    user_to_test["@404not_allowed_either"] = None
    user_to_test["@-illegal"] = None
    user_to_test["@not--allowed"] = None
    user_to_test["@illegal-2-"] = None
    user_to_test["@404Actually-allowed"] = "404Actually-allowed"
    user_to_test["@abcd"] = "abcd"
    user_to_test["@at/whats up"] = "at"
    user_to_test["@@bot"] = None
    user_to_test["@bothey@"] = "bothey"
    user_to_test["@mööüpasd"] = None
    user_to_test["@this_is_valit_until_#here"] = None
    user_to_test["@test.hey"] = "test"
    user_to_test["mail@example.com"] = None
    user_to_test["mail @example.com"] = "example"
    user_to_test["hey.@exple.How are you?I am fine!"] = "exple"  # yeah, that's a typo
    user_to_test["\@escaped_not_a_user"] = None

    def test_github_links(self):
        for input, output in self.urls_to_test.items():
            result = FILE_URL_REGEX.search(input)
            print ("\nInput   : {inp}\nexpected: {exp}\ngot     : {got}".format(inp=input, exp="MATCH" if output else "None", got=result))
            if output is None:
                self.assertIsNone(result, msg="Should not find anything.")
            else:
                self.assertIsNotNone(result, msg="Should find a url.")
            # end if
        # end for
    # end def

    def test_github_link_match_groups(self):
        input = "https://github.com/luckydonald/luckydonald-utils/blob/3888e7e6967e8b9b80245300321ddd2c19eff391/" \
                "luckydonaldUtils/eastereggs/headers.py#L12-L13"
        m = FILE_URL_REGEX.match(input)
        print(repr(m.groupdict()))
        self.assertEqual(m.group("user"),    "luckydonald", "user")
        self.assertEqual(m.group("repo"),    "luckydonald-utils", "repo")
        self.assertIn("path", m.groupdict(), "path")
        self.assertEqual(m.group("path"),    "blob/3888e7e6967e8b9b80245300321ddd2c19eff391/luckydonaldUtils/eastereggs/headers.py", "path")
        self.assertEqual(m.group("kind"),    "blob", "kind")
        self.assertEqual(m.group("branch"),  "3888e7e6967e8b9b80245300321ddd2c19eff391", "branch")
        self.assertEqual(m.group("file"),    "luckydonaldUtils/eastereggs/headers.py", "file")
        self.assertIn("hash", m.groupdict(), "hash")
        self.assertEqual(m.group("hash"),    "L12-L13", "hash")
    # end def

    def test_user_regex(self):
        for input, output in self.user_to_test.items():
            result = AT_USERNAME_REGEX.search(input)
            print ("\nInput   : {inp}\nexpected: {exp}\ngot     : {got}".format(inp=input, exp=output, got=result.group("user") if result else None))
            if output is None:
                self.assertIsNone(result, msg="Should not find anything.")
            else:
                self.assertEqual(output, result.group("user"), msg="Should find a @user.")
            # end if
        # end for
# end class




# -*- coding: utf-8 -*-
from setuptools import setup
from luckydonaldUtils import VERSION
from luckydonaldUtils.dependencies import find_submodules

__author__ = 'luckydonald'
long_description = """A collection of utilities I use across different projects"""

# http://peterdowns.com/posts/first-time-with-pypi.html
# $ python setup.py register -r pypi
# $ python setup.py sdist upload -r pypi

main_package = 'luckydonaldUtils'

install_requires = ["pip", "setuptools", "DictObject"]  # both should be installed by default.
try:
    import importlib
except ImportError:
    install_requires.append('importlib')

setup(
    name="luckydonald-utils",
    packages=find_submodules(main_package),
    version=VERSION,
    author="luckydonald",
    author_email="code_utils@luckydonald.de",
    description=long_description,
    license="GPLv2+",
    keywords="luckydonald utils utilities utility python dependencies dependency download progress bar encoding files "
             "interactions json update store text xml time network logger color images webserver django CSRF headers "
             "eastereggs decorator holder assert raise is None ponies pony",
    url="https://github.com/luckydonald/luckydonald-utils",
    install_requires=install_requires,
    long_description=long_description,
    # test_suite = 'nose.collector',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Utilities",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    ],
)



# -*- coding: utf-8 -*-
import re

__author__ = 'luckydonald'
__all__ = ['YOUTUBE_URL_REGEX', '_YOUTUBE_URL_REGEX']

_YOUTUBE_URL_REGEX = r'(?:https?://)?(?:www\.)?(?:youtube\.com/(?:v/|watch\?v=|embed/)|youtu\.be/)(?P<vid>[\w-]+)(?:\S+)?'
YOUTUBE_URL_REGEX = re.compile(_YOUTUBE_URL_REGEX)
"""
:const:`YOUTUBE_URL_REGEX`: Matches youtube videos.
The matching group `vid` contains the video id.
"""



# -*- coding: utf-8 -*-
# https://code.google.com/p/modwsgi/wiki/ReloadingSourceCode
"""
Add auto reloading capabilities.

Use in your wsgi file like this:

>>> import os
>>>
>>> from luckydonaldUtils.network.mod_wsgi import reloader
>>> reloader.start(interval=1.0)
>>> reloader.track(os.path.join(os.path.dirname(__file__), 'site.cf'))
>>>
>>> def application(environ, start_response):
>>>    ...

"""

from __future__ import print_function

__author__ = 'luckydonald'

from luckydonaldUtils.logger import logging  # pip install luckydonald-utils

logger = logging.getLogger(__name__)

import os
import sys
import signal
import threading
import atexit

try:
    from queue import Queue  # py3
except ImportError:
    from Queue import Queue  # py2
# end try

_interval = 1.0
_times = {}
_files = []

_running = False
_queue = Queue()
_lock = threading.Lock()


def warning(*objs):
    print(*objs, file=sys.stderr)


def _restart(path):
    _queue.put(True)
    prefix = 'monitor (pid=%d):' % os.getpid()
    warning('%s Change detected to \'%s\'.' % (prefix, path))
    warning('%s Triggering process restart.' % prefix)
    os.kill(os.getpid(), signal.SIGINT)


def _modified(path):
    try:
        # If path doesn't denote a file and were previously
        # tracking it, then it has been removed or the file type
        # has changed so force a restart. If not previously
        # tracking the file then we can ignore it as probably
        # pseudo reference such as when file extracted from a
        # collection of modules contained in a zip file.

        if not os.path.isfile(path):
            return path in _times

        # Check for when file last modified.

        mtime = os.stat(path).st_mtime
        if path not in _times:
            _times[path] = mtime

        # Force restart when modification time has changed, even
        # if time now older, as that could indicate older file
        # has been restored.

        if mtime != _times[path]:
            return True
    except:
        # If any exception occured, likely that file has been
        # been removed just before stat(), so force a restart.

        return True

    return False


def _monitor():
    while 1:
        # Check modification times on all files in sys.modules.

        for module in sys.modules.values():
            if not hasattr(module, '__file__'):
                continue
            path = getattr(module, '__file__')
            if not path:
                continue
            if os.path.splitext(path)[1] in ['.pyc', '.pyo', '.pyd']:
                path = path[:-1]
            if _modified(path):
                return _restart(path)

        # Check modification times on files which have
        # specifically been registered for monitoring.

        for path in _files:
            if _modified(path):
                return _restart(path)

        # Go to sleep for specified interval.

        try:
            return _queue.get(timeout=_interval)
        except:
            pass


_thread = threading.Thread(target=_monitor)
_thread.setDaemon(True)


def _exiting():
    try:
        _queue.put(True)
    except:
        pass
    _thread.join()


atexit.register(_exiting)


def track(path):
    if not path in _files:
        _files.append(path)


def start(interval=1.0):
    global _interval
    if interval < _interval:
        _interval = interval

    global _running
    _lock.acquire()
    if not _running:
        prefix = 'monitor (pid=%d):' % os.getpid()
        warning(sys.stderr, '%s Starting change monitor.' % prefix)
        _running = True
        _thread.start()
    _lock.release()
